--------------------------------------------------------------------------------
-- 
-- Name: WhoLib.lua
-- Author: Malex-Cenarius
-- Date: 12/23/2006
--
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
--
-- Method: RebuildCache
-- Description: 
--
--------------------------------------------------------------------------------
local function RebuildCache()

	WhoLibCache = {};
	WhoLibCache.Size = 0;
	WhoLibCache.Version = WhoLib.Constants.CacheVersion;

end

--------------------------------------------------------------------------------
--
-- Method: ClearPendingQuery
-- Description: 
--
--------------------------------------------------------------------------------
local LastQuery = "";
local function ClearPendingQuery()

	LastQuery = WhoLib.QueuePendingQuery;

	WhoLib.QueuePendingQuery = nil;
	WhoLib.QueuePendingElapsed = 0;		

end

--------------------------------------------------------------------------------
--
-- Method: RemoveQueueItem
-- Description: 
--
--------------------------------------------------------------------------------
local function RemoveQueueItem(itemName)

	WhoLib.Queue[itemName] = nil;
	WhoLib.QueueSize = WhoLib.QueueSize - 1;	

	WhoLib.Util.RemoveFilterFromQueue(itemName);
	
end

--------------------------------------------------------------------------------
--
-- Method: AddItemToCache
-- Description: 
--
--------------------------------------------------------------------------------
local function AddItemToCache(name, guild, level, race, class, zone)

	local nameLower = strlower(name);

	-- Add the user data to the cache
	--						
	WhoLibCache[nameLower] = {
		Guild = guild,
		Level = level,
		Race = race,
		Class = class,
		Zone = zone,
		Timestamp = time()
	};
	WhoLibCache.Size = WhoLibCache.Size + 1;
								
	local now = GetTime();
	local requestDuration = (now - WhoLib.Queue[nameLower].Timestamp) * 1000;			
	WhoLib.Util.DebugOut("|cFF00FF00COMPLETION: Adding user to cache: " .. name .. " CacheSize: (" .. WhoLibCache.Size .. ") Duration: " .. floor(requestDuration) .. " ms");
	
	-- Output to chat frame if necessary
	--	
	if(WhoLib.Queue[nameLower].OutputToChatFrame) then

		if(guild == "") then
			WhoLib.Util.Out(format("|cFFFFFF00" .. WHO_LIST_FORMAT, name, name, level, race, class, zone)); -- "|Hplayer:%s|h[%s]|h: Level %d %s %s - %s"; -- Example: Jeremy: Level 10 Human Warrior - Westfall		
		else		
			WhoLib.Util.Out(format("|cFFFFFF00" .. WHO_LIST_GUILD_FORMAT, name, name, level, race, class, guild, zone)); -- "|Hplayer:%s|h[%s]|h: Level %d %s %s <%s> - %s"; -- Example: Tigole: Level 40 Human Rogue <Legacy of Steel> - Westfall 
		end
		
		WhoLib.Util.Out(format("|cFFFFFF00" .. WHO_NUM_RESULTS, 1));
		
	end
	
	
	-- Remove the name from the queue
	--
	RemoveQueueItem(nameLower);
	
	-- Check to see if this result is also the pending request
	--
	if(WhoLib.QueuePendingIsComplex == false and WhoLib.QueuePendingQuery == nameLower) then
	
		-- Reset the pending request
		--
		ClearPendingQuery();
	
	end
	
	
	-- Check the total footprint of the cache
	--
	if(WhoLibCache.Size > WhoLib.Constants.MaxCachedEntries) then
	
		-- If the cache size has been exceeded, invalidate the entire
		-- cache.
		--
		RebuildCache();
	
	end

end

--------------------------------------------------------------------------------
--
-- Method: CheckWhoResults
-- Description: 
--
--------------------------------------------------------------------------------
local function CheckWhoResults()

	-- Check the who tab for results
	--
	for i = 1, GetNumWhoResults() do
	
		local name, guild, level, race, class, zone = GetWhoInfo(i);
		local nameLower = strlower(name);
				
		-- Check the queue for a matching result
		--
		if(WhoLib.Queue[nameLower] ~= nil) then				
			AddItemToCache(name, guild, level, race, class, zone);						
		end					
	
	end
	
	-- If the current query is complex, build up a results set
	--
	if(WhoLib.QueuePendingQuery ~= nil and WhoLib.QueuePendingIsComplex and GetNumWhoResults() > 0) then
	
		local now = GetTime();
		local results = { 
			Timestamp = now,
			Data = {}
		};					
		
		for i = 1, GetNumWhoResults() do
		
			local name, guild, level, race, class, zone = GetWhoInfo(i);
			
			local dataItem = {
				Name = name,
				Guild = guild,
				Level = level,
				Race = race,
				Class = class,
				Zone = zone,
				Timestamp = time()			
			};
			
			tinsert(results.Data, dataItem);
		
		end
		
		-- Add the results set to the results table
		--
		WhoLib.QueryResults[WhoLib.QueuePendingQuery] = results;
		WhoLib.QueryResults.n = WhoLib.QueryResults.n + 1;

		local requestDuration = (now - WhoLib.Queue[WhoLib.QueuePendingQuery].Timestamp) * 1000;						
		WhoLib.Util.DebugOut("|cFF00FF00COMPLETION: Adding result table: " .. WhoLib.QueuePendingQuery .. " (" .. getn(results.Data) .. ") Duration: " .. floor(requestDuration) .. " ms");			

		-- Remove the query from the queue
		--
		RemoveQueueItem(WhoLib.QueuePendingQuery);
		
		-- Reset the pending request
		--
		ClearPendingQuery();	
		
	end

end

--------------------------------------------------------------------------------
--
-- Method: ProcessQueue
-- Description: 
-- Parameters:
--	elapsed - [in] The elapsed time in milliseconds
--
--------------------------------------------------------------------------------
local function ProcessQueue(elapsed)	
			
	-- Update the current who request if it exists
	--
	if(WhoLib.QueuePendingQuery ~= nil) then
	
		-- Add the elapsed time to the current pending request
		--
		WhoLib.QueuePendingElapsed = WhoLib.QueuePendingElapsed + elapsed;
		
		-- Timeout the request if the elapsed time exceeds the timeout
		--
		if(WhoLib.QueuePendingElapsed >= WhoLib.Constants.WhoRequestTimeout) then
				
			-- Remove the item from the queue
			--
			if(WhoLib.Queue[WhoLib.QueuePendingQuery].Retries <= 0) then
			
				-- Output a null search results text to the chat frame
				--
				if(WhoLib.Queue[WhoLib.QueuePendingQuery].OutputToChatFrame) then				
					WhoLib.Util.Out(format("|cFFFFFF00" .. WHO_NUM_RESULTS, 0));
				end
			
				-- Remove the item from the queue
				--				
				RemoveQueueItem(WhoLib.QueuePendingQuery);

				-- If a complex query timed out, add an empty results table 
				--
				if(WhoLib.QueuePendingIsComplex) then
				
					WhoLib.Util.DebugOut("|cFFFF0000TIMEOUT: query timed out.  Adding empty results table for: " .. WhoLib.QueuePendingQuery .. " (" .. WhoLib.QueueSize .. ")");
				
					local results = { 						
						Timestamp = GetTime(),
						Data = {}
					};								
				
					WhoLib.QueryResults[WhoLib.QueuePendingQuery] = results;
					WhoLib.QueryResults.n = WhoLib.QueryResults.n + 1;
								
				end				
			
			else

				-- Decrease the retries remaining
				--
				WhoLib.Queue[WhoLib.QueuePendingQuery].Retries = WhoLib.Queue[WhoLib.QueuePendingQuery].Retries - 1;
			
			end			
			
			-- Reset the current pending item
			--
			ClearPendingQuery();					
					
		end
	
	end
	
	-- Start a new who request if there isn't already one pending, there are requests in the queue, if 
	-- the user isn't currently using the who popup, and if there has been sufficient time since the
	-- last user-initiated who request
	--	
	if(WhoLib.QueuePendingQuery == nil and WhoLib.QueueSize > 0 and WhoLib.FriendsFrameIsVisible == false) then
		
		-- Grab the next entry in the queue
		--
		if(getn(WhoLib.QueueFilterList) > 0) then
		
			local query = WhoLib.QueueFilterList[1];
			local value = WhoLib.Queue[query];
						
			if(value ~= nil) then
					
				-- Check to make sure the request didn't timeout while in the queue
				--
				if(GetTime() - value.Timestamp >= WhoLib.Constants.WhoQueueTimeout) then			
				
					-- Remove the item from the queue
					--
					RemoveQueueItem(query);					
					
					WhoLib.Util.DebugOut("|cFFFF0000Invalidating Request: " .. query .. " (" .. WhoLib.QueueSize .. ")");
				
				else					
			
					WhoLib.Util.DebugOut("|cFF0080FFExecuting Request: " .. query);
				
					-- Set the current request
					--
					WhoLib.QueuePendingQuery = query;
					WhoLib.QueuePendingElapsed = 0;
					WhoLib.QueuePendingIsComplex = value.IsComplex;
					
					-- Execute the request
					--			
					SetWhoToUI(1);
					WhoFrameEditBox:SetText(query);
					SendWho(query);
				
				end				
			
			end
		
		end
		
	end

end

--------------------------------------------------------------------------------
--
-- Method: LoadSavedVariables
-- Description: 
--
--------------------------------------------------------------------------------
local function LoadSavedVariables()

	-- Create a new cache table if it doesn't exist or if the format has changed
	--
	if(WhoLib.Constants.IsDebug or WhoLibCache == nil or WhoLibCache.Version == nil or WhoLibCache.Version ~= WhoLib.Constants.CacheVersion) then

		RebuildCache();		
	
	end

end

--------------------------------------------------------------------------------
--
-- Method: InvalidateQueryResults
-- Description: This method gets executed when a user-initiated who query
--	completes.  In this case, if there are any background queries running, 
--	invalidate the results.
--
--------------------------------------------------------------------------------
local function InvalidateQueryResults()

	-- Immediately timeout any pending request
	--
	ClearPendingQuery();

end

--------------------------------------------------------------------------------
--
-- Method: TimeoutQueryResults
-- Description: This method gets called to remove query results that haven't
--	been retrieved and have hit the timeout
--
--------------------------------------------------------------------------------
function TimeoutQueryResults()
	
	for key, value in pairs(WhoLib.QueryResults) do
	
		-- Check each results set to see the timeout has expired
		--
		if(type(value) == "table" and GetTime() - value.Timestamp > WhoLib.Constants.QueryResultsTimeout) then
	
			WhoLib.Util.DebugOut("Timing out results table: " .. key);
	
			-- Remove the results from the results table
			--
			WhoLib.QueryResults[key] = nil;
			WhoLib.QueryResults.n = WhoLib.QueryResults.n - 1;
		
		end
	
	end

end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_ChatFrame_OnEvent_Override
-- Description: Override for the chat frame in order to remove who response
--	system messages for background queries
--
--------------------------------------------------------------------------------
function WhoLib_ChatFrame_OnEvent_Override(...)			
		
	if(event == "CHAT_MSG_SYSTEM" and WhoLib.QueuePendingQuery ~= nil and (string.find(arg1, WhoLib.QueuePendingQuery) ~= nil or WhoLib.Util.ContainsSystemWhoText(arg1))) then
	
		-- Some who request has completed, so let the update handler run
		--
		if(WhoLib.Util.ContainsSystemWhoText(arg1)) then
			ProcessQueue(0);
		end							
		
	else
			
		WhoLib.ChatFrame_OnEvent_Original(...);			
	
	end		

end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_ChatEdit_OnEnterPressed_Override
-- Description: Override for the chat editbox in order to determine when a 
--	user-initiated /who request starts
--
--------------------------------------------------------------------------------
function WhoLib_ChatEdit_OnEnterPressed_Override()
	
	local text = this:GetText();
	local textLen = strlen(text);
	local who1Len = strlen(SLASH_WHO1);	
	
	if(string.find(strlower(text), SLASH_WHO1) == 1 and string.find(text, " ") == who1Len + 1 and textLen > who1Len + 1) then
	
		local whoName = strsub(text, who1Len + 2);
	
		WhoLib.Util.DebugOut("user-initiated /who command intercepted for: " .. whoName);

		InvalidateQueryResults();
		
		-- Add the who request to the queue
		--
		WhoLib_GetUserInfo(whoName, 0, true, true, 2);
			
	end	

end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_FriendsFrame_OnEvent_Override
-- Description: Override for the friends frame in order to avoid showing the
--	popup during complex queries
--
--------------------------------------------------------------------------------
function WhoLib_FriendsFrame_OnEvent_Override(...)
		
	if(event == "WHO_LIST_UPDATE") then
	
		local whoFilter = WhoFrameEditBox:GetText();		

		if(whoFilter == LastQuery or whoFilter == WhoLib.QueuePendingQuery) then
			
			CheckWhoResults();
						
			if(WhoLib.FriendsFrameIsVisible == false) then
				return;
			end
									
		end

	end
	
	WhoLib.FriendsFrame_OnEvent_Original(...);

end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_FriendsFrame_OnShow_Override
-- Description: 
--------------------------------------------------------------------------------
function WhoLib_FriendsFrame_OnShow_Override(...)
	
	WhoLib.FriendsFrameIsVisible = true;
	WhoLib.FriendsFrame_OnShow_Original(...);

end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_FriendsFrame_OnHide_Override
-- Description: 
--
--------------------------------------------------------------------------------
function WhoLib_FriendsFrame_OnHide_Override(...)
	
	WhoLib.FriendsFrameIsVisible = false;
	WhoLib.FriendsFrame_OnHide_Original(...);

end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_OnLoad
-- Description: Called during the onload of the addon
--
--------------------------------------------------------------------------------
function WhoLib_OnLoad()
	
	-- Event Registration
	--
    this:RegisterEvent("VARIABLES_LOADED");
	this:RegisterEvent("WHO_LIST_UPDATE");	
    	
	-- Hook the chat frame
	--
	WhoLib.ChatFrame_OnEvent_Original = ChatFrame_OnEvent;
	ChatFrame_OnEvent = WhoLib_ChatFrame_OnEvent_Override;	

	-- Hook the chat editbox
	--
	--WhoLib.ChatEdit_OnEnterPressed_Original = ChatEdit_OnEnterPressed;	
	hooksecurefunc("ChatEdit_OnEnterPressed", WhoLib_ChatEdit_OnEnterPressed_Override);	
	--ChatEdit_OnEnterPressed = WhoLib_ChatEdit_OnEnterPressed_Override;
	
	-- Hook the friends frame
	--
	WhoLib.FriendsFrame_OnEvent_Original = FriendsFrame_OnEvent;
	FriendsFrame_OnEvent = WhoLib_FriendsFrame_OnEvent_Override;
	
	WhoLib.FriendsFrame_OnShow_Original = FriendsFrame_OnShow;
	FriendsFrame_OnShow = WhoLib_FriendsFrame_OnShow_Override;
	
	WhoLib.FriendsFrame_OnHide_Original = FriendsFrame_OnHide;
	FriendsFrame_OnHide = WhoLib_FriendsFrame_OnHide_Override;
	
		
	if(WhoLib.Constants.UseTestData) then
		this:RegisterEvent("CHAT_MSG_CHANNEL");
		this:RegisterEvent("CHAT_MSG_WHISPER");	    
	    this:RegisterEvent("CHAT_MSG_YELL");	
		this:RegisterEvent("CHAT_MSG_GUILD");
	end

end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_OnUpdate
-- Description: Called every frame while the addon is being shown
-- Parameters:
--	elapsed - [in] The amount of time in seconds since the last call to OnUpdate
--
--------------------------------------------------------------------------------
local timeSinceLastProcessQueue = WhoLib.Constants.ProcessQueueFrequency;
local timeSinceLastTimeoutQueryResults = 0;
local debugQueryIndex = 1;
function WhoLib_OnUpdate(elapsed)

	-- Service the query queue
	--
	timeSinceLastProcessQueue = timeSinceLastProcessQueue + (elapsed * 1000.0);
	if(timeSinceLastProcessQueue >= WhoLib.Constants.ProcessQueueFrequency) then				
	
		ProcessQueue(timeSinceLastProcessQueue);				
		timeSinceLastProcessQueue = 0.0;
	
	end
	
	-- Timeout any results data that hasn't been retrieved yet
	--
	timeSinceLastTimeoutQueryResults = timeSinceLastTimeoutQueryResults + elapsed;
	if(timeSinceLastTimeoutQueryResults >= WhoLib.Constants.TimeoutQueryResultsFrequency) then
	
		if(WhoLib.Constants.IsDebug and WhoLib.Constants.TestQuery) then
			
			debugQueryIndex = mod(debugQueryIndex, 2) + 1;
			local queryIndex = debugQueryIndex; --math.random(1, 5);								
			local isValid, data = WhoLib_Query(WhoLib.DebugQueries[queryIndex]);
		
			if(isValid) then
			
				WhoLib.Util.DebugOut("Query results retrieved containing: " .. getn(data) .. " entries for query: " .. WhoLib.DebugQueries[queryIndex]);
			
				-- Validate data
				--
				for index, value in ipairs(data) do				
													
					if(strfind(WhoLib.DebugQueries[queryIndex], value.Class) == nil) then
						
						WhoLib.Util.DebugOut("|cFFFF0000Invalid data received for query: " .. WhoLib.DebugQueries[queryIndex] .. " found class: " .. value.Class);
						break;
					
					end
				
				end
			
			end
		
		end
	
		TimeoutQueryResults();
		timeSinceLastTimeoutQueryResults = 0.0;
	end
	
end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_OnEvent
-- Description: The event handles for the addon
--
--------------------------------------------------------------------------------
function WhoLib_OnEvent(event)	
						
	if(event == "VARIABLES_LOADED") then		
		
		-- Load saved variables
		--
		LoadSavedVariables();		
	
	elseif(event == "WHO_LIST_UPDATE") then
						
		-- This event signals that the user has executed a who query in the who popup
		-- If there are any queries pending, invalidate their results
		--		
		if(WhoLib.Util.IsWhoFrameVisible()) then
		
			WhoLib.Util.DebugOut("|cFFFFFF00Detecting user initiated who request completed.");	

			-- Reset the retry count for the next item in the queue
			--
			if(getn(WhoLib.QueueFilterList) > 0) then
			
				local value = WhoLib.Queue[WhoLib.QueueFilterList[1]];
				value.Retries = WhoLib.Constants.MaxRetries;
			
			end			
		
			if(WhoLib.QueuePendingQuery ~= nil) then

				WhoLib.Util.DebugOut("|cFFFFFF00Detecting user initiated who request completed while query pending.");			
				InvalidateQueryResults();		
				
			end
		
		end
		
	elseif(event == "CHAT_MSG_CHANNEL" or event == "CHAT_MSG_WHISPER" or event == "CHAT_MSG_YELL" or event == "CHAT_MSG_GUILD") then
			
		local isValid, guildName, level, race, class, zone, timestamp = WhoLib_GetUserInfo(arg2);
		if(isValid) then
		
			local secondsPassed = time() - timestamp;
			local friendlyTimePassed = SecondsToTime(secondsPassed);
			
			WhoLib.Util.DebugOut("UserDataFound: Name:" .. arg2 .. " " .. guildName .. " " .. level .. " " .. race .. " " .. class .. " " .. zone .. " " .. friendlyTimePassed);
		end
	
	end	

end


WhoLib.Util.ProcessQueue = ProcessQueue;